'''uses methods from the ``gini`` library for flexible matching'''
from gini.semantics import Bottle,Concept
from gini import logic
import padre as p
import neural as nl
import copy

bottle = Bottle()

p.subjects()
bottle.vocab = [
    # Direct objects
    Concept('atlas_do', ['atlases','templates']),
    Concept('subject_do', ['subjects','patients']),
    Concept('experiment_do', ['experiments', 'studies', 'study']),
    Concept('label_do', ['labels','tasks']),
    Concept('tag_do', ['types','tags']),
    Concept('session_do', ['sessions','scans','dates']),
    Concept('dset_do', ['dsets','datasets','runs','files']),
    Concept('metadata_do', ['meta','metadata','behavioral','eprime']),
    Concept('eprime_do', ['eprime','edat']),
    Concept('eprime-txt_do', ['eprime-txt']),
    # Actions
    Concept('list', ['list','show','get','print']),
    Concept('link', ['link','copy','symlink']),
    Concept('add', ['add','attach','upload','import']),
    Concept('new', ['new','create']),
    # Objects
    Concept('subject', [[x] for x in p.subject._all_subjects.keys()]),
    Concept('label', [[x] for x in p.subject.tasks]),
    Concept('tag', [[x] for x in p.subject.tags]),
    Concept('experiment', [[x] for x in p.subject.experiments]),
    # Options
    Concept('quiet', ['quiet','quietly','-q','--quiet']),
    Concept('all',['all','include','incomplete','everything'])
]

def bottle_help():
    nl.notify('''This program uses freeform text fuzzy matching. It will try to correct for misspellings and
    synonyms as it can.
    
    General syntax:
    [action] [objects]
    
    Currently supported actions are:
        list    print something to the screen (add quiet to do it less loud)
        add     add on something to an existing subject
        new     create a new subject
        link    create symbolic links to datasets in current directory
    
    Examples:
        list subjects
        list subjects ExperimentName
        get dsets SubjectName
        get subjects LabelName Tag1
    
    Just try stuff...

    ''')


def recursive_parse_string(bottle,string):
    raise RuntimeError('Need to update to newest gini version')
    '''calls :meth:`bottle.parse_string` and then adds subject-specific information and re-runs it to get things like dataset names'''
    matches = bottle.parse_string(string)
    if 'other' in matches:
        # There were some strings that didn't match anything
        matching_subjects = filter_subjs(p.subjects(),matches)
        new_bottle = copy.deepcopy(bottle)
        new_bottle.vocab['dset'] = []
        # Not finished -- add dset filenames, metadata, 
        # Would be nice to keep track of where they came from (what subject is this dset from?)

def dsets_with(subj,args):
    raise RuntimeError('Need to update to newest gini version')
    '''returns the result of :meth:`Subject.dsets`, using the relevant keyword arguments from args'''
    def get(dict,key,only_one=True):
        if key in dict:
            if only_one and isinstance(dict[key],list):
                return dict[key][0]
            else:
                return dict[key]
        else:
            return None
    return subj.dsets(label=get(args,'label'),tags=get(args,'tag',False),experiment=get(args,'experiment'))

def filter_by_match(match,subjects):
    if match.concept.name=='subject':
        # Combine subjects by OR (AND doesn't make any sense..)
        return [x for x in subjects if str(x)==match.examples]
    if match.concept.name=='experiment':
        # Combine experiments by OR
        filtered = []
        for subj in subjects:
            if any([subj._sessions[sess]['experiment'] in match.examples for sess in subj._sessions if 'experiment' in subj._sessions[sess] and subj._sessions[sess]['experiment']]):
                filtered.append(subj)
        return filtered
    if match.concept.name=='label':
        # Combine with AND per subject (not necessarily the same session)
        filtered = []
        for subj in subjects:
            if any([any([label in match.examples for label in subj._sessions[sess]['labels']]) for sess in subj._sessions]):
                filtered.append(subj)
        return filtered
    if match.concept.name=='tag':
        # Combine with AND per subject (not necessarily the same session)
        filtered = []
        for subj in subjects:
            if any([any([tag in match.examples for tag in subj._sessions[sess]['tags']]) for sess in subj._sessions if 'tags' in subj._sessions[sess]]):
                filtered.append(subj)
        return filtered

def filter_subjs(subjects=None,string=None,matches=None,require_match=True):
    '''takes ``list`` of subjects and filters based on the :class:`ConceptMatch`s given in ``matches``. If ``matches``
    is not given, will parse the string ``string`` instead. If ``require_match``, will return ``None`` if it fails to find
    any constraints, otherwise it returns ``subjects``'''
    match_concepts = ['subject','session','label','experiment','tag']
    if matches==None:
        if string==None:
            if subjects==None:
                return p.subjects()
            else:
                return subjects
        matches = bottle.parse_string(string)
    
    if subjects==None:
        if 'all' in [x[0].concept.name for x in matches]:
            nl.notify('Including incomplete sessions')
            subjects = p.subjects(only_included=False)
        else:
            subjects = p.subjects()
    running_exclusions = {}
    if not any([[x in sub_matches for x in match_concepts] for sub_matches in matches]):
        if require_match:
            nl.notify('Warning: Could not find any constraints to use',level=nl.level.warning)
            return None
        else:
            nl.notify('Using no constraints')
            return subjects
    with nl.notify('Using constraints:'):
        subjects_total = set(subjects)
        for match_options in matches:
            if not isinstance(match_options,list):
                continue
            match_options = [x for x in match_options if x in match_concepts]
            if len(match_options)==0:
                continue
            subjects_options = set()
            for match in match_options:
                nl.notify('%s = %s' % (match.concept.name,repr(match.examples)))
                subjects_options = logic.or_sets(subjects_options,set(filter_by_match(match,subjects)))
            subjects_total = logic.and_sets(subjects_options,subjects_total)
    return subjects_total



#!/usr/bin/env python
# depends on:
#   neural
#   padre
#   fuzzywuzzy
#   openpyxl

import argparse,os,json,sys,shutil,getpass
from fuzzywuzzy import process
import padre as p
import neural as nl
import openpyxl
from padre.matching import bottle,filter_subjs,dsets_with,bottle_help


def list_objects(args):
    with nl.notify('Listing objects I can find that match that...'):
        subjects = filter_subjs(matches=args)
        
        dos = [x[0].concept.name for x in args if 'concept' in dir(x[0]) and x[0].concept.name.endswith('_do')]
        if 'subject_do' in dos:
            nl.notify('Subjects:')
            print '\n'.join(subjects)
    
        session_dicts = [x._sessions for x in subjects]
    
        if 'session_do' in dos:
            nl.notify('Sessions:')
            print '\n'.join(set([x.keys() for x in session_dicts]))

        sessions = [a for b in [x.values() for x in session_dicts] for a in b]
    
        if 'label_do' in dos:
            nl.notify('Labels:')
            labels = set([a for b in [x['labels'].keys() for x in sessions] for a in b])
            print '\n'.join(labels)
    
        if 'tag_do' in dos:
            nl.notify('Tags:')
            tags = set([a for b in [x['tags'] for x in sessions if 'tags' in x] for b in a])
            print '\n'.join(tags)
    
        if 'experiment_do' in dos:
            nl.notify('Experiments:')
            experiments = set([x['experiment'] for x in sessions if 'experiment' in x and x['experiment']])
            print '\n'.join(experiments)

def link_dsets(args):
    with nl.notify('Trying to link the following datasets...'):
        subjects = filter_subjs(p.subjects(),args)
        dsets_to_link = []
        
        for subj in subjects:
            with nl.notify(str(subj)):
                for dset in dsets_with(subj,args):
                    nl.notify(dset.__str__(False))
                    dsets_to_link.append(dset)
        nl.notify('Does that all look good? (y/n)')
        i = raw_input()
        if i.lower()=='y':
            nl.notify('Ok, linking them')
            for dset in dsets_to_link:
                try:
                    os.symlink(str(dset),os.path.basename(str(dset)))
                except OSError:
                    pass
        else:
            nl.notify('Nope, that didn\'t look like a "y"...',level=nl.level.warning)

def new_object(args):
    args = [x for x in args if 'concept' not in dir(x[0]) or x[0].concept.name!='new']
    concept_names = [x[0].concept.name for x in args if 'concept' in dir(x[0])]
    if 'session_do' in concept_names:
        other_args = [x for x in args if isinstance(x,basestring)]
        if len(other_args)>1:
            nl.notify('Error: You gave me multiple names (%s), don\'t know which you want to be the session name!' % ','.join(other_args),level=nl.level.error)
            return False
        new_session = other_args[0]
        if 'subject' not in concept_names:
            nl.notify('Error: I don\'t know what subject to create session %s in!' % new_session,level=nl.level.error)
            return False
        subject_id = [x[0].item for x in args if 'concept' in dir(x[0]) and x[0].concept.name=='subject'][0]
        with nl.notify('Creating session %s in subject %s' % (new_session,subject_id)):
            subj = p.load(subject_id)
            if not subj:
                nl.notify('Error: couldn\'t load data for subject "%s"' % subject_id,level=nl.level.error)
                return False
            p.maint.new_session(subj,new_session)
            return True
    # Otherwise assume you want to make a subject
    other_args = [x for x in args if isinstance(x,basestring)]
    if len(other_args)>1:
        nl.notify('Error: You gave me multiple names (%s), don\'t know which you want to be the subject name!' % ','.join(other_args),level=nl.level.error)
        return False
    new_subject = other_args[0]
    with nl.notify('Creating new subject %s' % (new_subject)):
        p.maint.create_subject(new_subject)
        return True

def add_data(args):
    pass

def error(msg,miss=None):
    nl.notify(msg + '\n',level=nl.level.error)
    if miss:
        try:
            import urllib,urllib2
            data = {'miss':miss}
            data_enc = urllib.urlencode(data)
            urllib2.urlopen('http://wolflion.org/cgi-bin/report.py?%s' % data_enc)
        except:
            pass
    sys.exit()

'''def padre_add_meta(args):
    (params,listable,unidentified_args) = identify_params(args,True)
    all_subjs = subjects_from_params(params)
    (all_dsets,unidentified_args) = identify_dsets(unidentified_args,all_subjs)
    if len(all_dsets)!=1:
        error('Error: To add meta-data, you need to give me the name of exactly one dataset')
    (subj,sess,label,dset,i) = all_dsets.values()[0]
    if 'meta' not in subj._sessions[sess]['labels'][label][i]:
        subj._sessions[sess]['labels'][label][i]['meta'] = {}
    subj._sessions[sess]['labels'][label][i]['meta'][args.meta_type] = os.path.basename(args.filename)
    subj.save()
    shutil.copy(args.filename,os.path.join(p.sessions_dir(subj),sess))
    print 'Added %s (%s) to %s' % (os.path.basename(args.filename),args.meta_type,str(subj))'''
        

bottle.set_action('list',list_objects)
bottle.set_action('link',link_dsets)
bottle.set_action('new',new_object)
bottle.set_action('add',add_data)

if __name__ == '__main__':
    if len(sys.argv)==1:
        bottle_help()
    else:
        bottle.process_string(sys.argv[1:])